if ( not MarketWatcher ) then
	return
end

local MarketWatcher = MarketWatcher;
local timer = MarketWatcher.timer;

if ( not MarketWatcher.Scan ) then
	MarketWatcher.Scan = {};
end
local scan = MarketWatcher.Scan;

scan.status = {};
local status = scan.status;
scan.queries = {};
local queries = scan.queries;
scan.scanStats = {};
local scanStats = scan.scanStats;

StaticPopupDialogs["MW_BUYOUT_AUCTION"] = {
	text = BUYOUT_AUCTION_CONFIRMATION,
	button1 = ACCEPT,
	button2 = CANCEL,
	OnAccept = function(self)
		local _, _, _, _, _, _, _, _, buyoutPrice = GetAuctionItemInfo("list", GetSelectedAuctionItem("list"));
		PlaceAuctionBid("list", GetSelectedAuctionItem("list"), buyoutPrice);
	end,
	OnShow = function(self)
		local _, _, _, _, _, _, _, _, buyoutPrice = GetAuctionItemInfo("list", GetSelectedAuctionItem("list"));
		MoneyFrame_Update(self.moneyFrame, buyoutPrice);
	end,
	hasMoneyFrame = 1,
	showAlert = 1,
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
};

StaticPopupDialogs["MW_RECORD_CONFIRM"] = {
	text = MARKETWATCHER_RECORD_CONFIRM_POPUP,
	button1 = YES,
	button2 = NO,
	OnAccept = function(self)
		MarketWatcherScanFrameRecordCheckBox:SetChecked(1);
	end,
	showAlert = 1,
	timeout = 0,
	exclusive = 0,
	hideOnEscape = 1
};

---------------------------------------------------------------------------------
--	Tab Functions
---------------------------------------------------------------------------------

function scan.SetScanTab()

	scan.SetupHook();

	local tabs = 1;
	while ( _G["AuctionFrameTab"..tabs] ) do
		if ( _G["AuctionFrameTab"..tabs]:GetName() == "AuctionFrameTabMarketWatcherScan" ) then
			return
		end
		tabs = tabs + 1;
	end

	local button = CreateFrame("Button", "AuctionFrameTabMarketWatcherScan", AuctionFrame, "AuctionTabTemplate");
	button:SetText(MARKETWATCHER_SCAN);
	PanelTemplates_TabResize(button, 0);
	PanelTemplates_DeselectTab(button);

	_G["AuctionFrameTab"..tabs] = AuctionFrameTabMarketWatcherScan;
	button = _G["AuctionFrameTab"..tabs];
	button:SetParent("AuctionFrame");
	button:SetPoint("TOPLEFT", _G["AuctionFrameTab"..(tabs - 1)]:GetName(), "TOPRIGHT", -8, 0);
	button:SetID(tabs);
	button:Show();

	PanelTemplates_SetNumTabs(AuctionFrame, tabs);

	MarketWatcherScanFrame:SetParent(AuctionFrame);
	MarketWatcherScanFrame:SetPoint("TOPLEFT", AuctionFrame, "TOPLEFT", 0, 0);

	hooksecurefunc("AuctionFrameTab_OnClick", scan.Tab_OnClick);
	button:SetScript("OnClick", AuctionFrameTab_OnClick);
end

function scan.Tab_OnClick(self)
	if ( not self ) then
		MarketWatcherScanFrame:Hide();
		return
	end
	local index = self:GetID();
	local tab = getglobal("AuctionFrameTab"..index);
	if ( tab and tab:GetName() == "AuctionFrameTabMarketWatcherScan" ) then
		PanelTemplates_SetTab(AuctionFrame, index);

		AuctionFrameTopLeft:SetTexture("Interface\\AuctionFrame\\UI-AuctionFrame-Bid-TopLeft");
		AuctionFrameTop:SetTexture("Interface\\AuctionFrame\\UI-AuctionFrame-Bid-Top");
		AuctionFrameTopRight:SetTexture("Interface\\AuctionFrame\\UI-AuctionFrame-Bid-TopRight");
		AuctionFrameBotLeft:SetTexture("Interface\\AuctionFrame\\UI-AuctionFrame-Bid-BotLeft");
		AuctionFrameBot:SetTexture("Interface\\AuctionFrame\\UI-AuctionFrame-Bid-Bot");
		AuctionFrameBotRight:SetTexture("Interface\\AuctionFrame\\UI-AuctionFrame-Bid-BotRight");

		MarketWatcherScanFrame:Show();
	else
		MarketWatcherScanFrame:Hide();
	end
end

-- hook the setup function
scan.SetupHook = MarketWatcher.Setup;
MarketWatcher.Setup = scan.SetScanTab;


---------------------------------------------------------------------------------
--	Scanning Functions
---------------------------------------------------------------------------------

function scan.RecordCheckBox_OnShow(self)
	if ( MarketWatcher.GetConfig().lastRecordedScan and MarketWatcher.GetConfig().lastRecordedScan > (time() - 3600) ) then
		self:SetChecked(false);
	else
		self:SetChecked(1);
	end
end

function scan.RecordCheckBox_OnClick(self)
	if ( self:GetChecked() ) then
		if ( MarketWatcher.GetConfig().lastRecordedScan and MarketWatcher.GetConfig().lastRecordedScan > (time() - 3600) ) then
			self:SetChecked(false);
			StaticPopup_Show("MW_RECORD_CONFIRM");
		end
	end
end

function scan.Scan()
	status.scan = true;
	status.query = 1;
	status.page = 0;
	status.maxPages = 0;
	status.pagesScanned = 0;
	status.queryTime = time(date("*t"));
	status.startTime = time();
	scan.processed = {};

	scan.selectedItem = nil;

	MarketWatcherScanFrameStatusText:Show();
	MarketWatcherScanFrameStatusText2:Show();
	MarketWatcherScanFrameAbortButton:Show();
	MarketWatcherScanFrameScanTimeText:SetText("");

	local i = 1;

	while ( _G["MarketWatcherScanFrameStatsButton"..i] ) do
		_G["MarketWatcherScanFrameStatsButton"..i]:Hide();
		i = i + 1;
	end
	MarketWatcherScanFrameStatsScrollFrame:Hide();
	MarketWatcherScanFrameAuctions:Hide();

	-- empty tables
	for i in ipairs(queries) do
		queries[i] = nil;
	end
	for i in ipairs(scanStats) do
		scanStats[i] = nil;
	end
	tempItemTable = MarketWatcher.NewTempTable();

	i = false;

	-- create empty tables for itemIds so we can track items with zero auctions
	for itemId, itemData in MarketWatcher.WatchedItemsIterator() do
		if ( itemData.name ) then
			MarketWatcher.AddTempItem(itemId, status.queryTime);
		end
		i = true;
	end

	-- any watched items?
	if ( not i ) then
		scan.Abort();
		MarketWatcherScanFrameStatusText:Show();
		MarketWatcherScanFrameStatusText:SetText(MARKETWATCHER_NOWATCHEDITEMS);
		return
	end

	queries = MarketWatcher.GenerateQueryList();

	MarketWatcher.listFunction = scan.ListUpdate;

	if ( #queries > 0 ) then
		scan.QueryServer();
	end
end

function scan.Abort()
	MarketWatcher.listFunction = nil;
	timer.StopTimer();

	MarketWatcherScanFrameAbortButton:Hide();
	MarketWatcherScanFrameScanButton:Enable();

	status.scan = false;
	MarketWatcherScanFrameStatusText:Hide();
	MarketWatcherScanFrameStatusText2:Hide();

end

function scan.Close()
	
	scan.CloseHook();

	if ( scan.status.scan ) then
		-- scan incomplete
		scan.Abort();
	end

	scan.itemSelected = nil;
	MarketWatcherScanFrameAuctions:Hide();
end

scan.CloseHook = MarketWatcher.Close;
MarketWatcher.Close = scan.Close;

function scan.QueryServer()

	if ( CanSendAuctionQuery() ) then

		local query = queries[status.query];
		local s, class, subclass, qualityText;

		if ( status.page == 0 ) then
			s = format(MARKETWATCHER_PAGESTATUS, status.page + 1);
		else
			s = format(MARKETWATCHER_PAGESTATUS2, status.page + 1, status.maxPages + 1);
		end
		MarketWatcherScanFrameStatusText:SetText( format(MARKETWATCHER_QUERYSTATUS, status.query, #queries).."\n"..s );

		status.waitForNames = false;
		status.queryTime = time(date("*t"));

		class = MarketWatcher.CLASSES[query.classIndex] or "";
		subclass = MarketWatcher.SUBCLASSES[query.classIndex][query.subclassIndex] or "";
		qualityText = _G["ITEM_QUALITY"..query.qualityIndex.."_DESC"] or "";

		MarketWatcher.QueryServer(query.name, query.minLevel, query.maxLevel, query.invTypeIndex, query.classIndex, query.subclassIndex, status.page, 0, query.qualityIndex, false);

		MarketWatcherScanFrameStatusText2:SetText( format(MARKETWATCHER_QUERYSTATUS2, query.name or "", query.invTypeIndex or "", class, subclass, qualityText, status.page) );
		if ( MarketWatcher.debug ) then
			print( format(MARKETWATCHER_QUERYSTATUS2, query.name or "", query.invTypeIndex or "", class, subclass, qualityText, status.page) );
		end
	else
		timer.StartTimer(.1, scan.QueryServer);
	end
end

function scan.ListUpdate()
	if ( not status.scan ) then
		return
	end

	local batch, totalAuctions = GetNumAuctionItems("list");
	if ( totalAuctions > 0 ) then
		status.maxPages = ceil(totalAuctions / NUM_AUCTION_ITEMS_PER_PAGE) - 1;
	else
		status.maxPages = 0;
	end

	-- are we waiting for seller names?
	if ( status.waitForNames ) then

		-- get names, and complete page if all names are in
		if ( not scan.GetSellerNames() ) then

			-- refresh timer
			timer.StartTimer(15, scan.NextPage);

			MarketWatcherScanFrameStatusText2:SetText(MARKETWATCHER_LISTUPDATE3);

			return
		end
	else

		scan.GetAuctions();

		-- complete page if auctions all had seller names
		if ( status.waitForNames ) then

			-- start timer in case of nil owners
			timer.StartTimer(15, scan.NextPage);

			MarketWatcherScanFrameStatusText2:SetText(MARKETWATCHER_LISTUPDATE2);

			return
		end
	end

	-- if we are here, then ready for next page/query

	status.pagesScanned = status.pagesScanned + 1;
	MarketWatcherScanFrameStatusText2:SetText(MARKETWATCHER_LISTUPDATE1);

	scan.NextPage();
end

function scan.NextPage()
	scan.OnPageComplete();

	if ( status.page < status.maxPages ) then
		status.page = status.page + 1;
	else
		status.query = status.query + 1;
		status.page = 0;

		scan.OnQueryComplete();

		if ( not queries[status.query] ) then

			-- scan completed
			scan.OnScanComplete();

			return
		end
	end

	timer.StartTimer(MarketWatcher.GetConfig().queryDelay, scan.QueryServer);
end

function scan.GetAuctions()

	local batch, total = GetNumAuctionItems("list");
	local index;

	local name, texture, count, quality, canUse, level, minBid, minIncrement, buyoutPrice, bidAmount, highBidder, owner;
	local timeleft, itemLink, itemId, itemMods, uId;
	--local itemLevel, itemType, itemSubType, itemEquipLoc;

	for index = 1, batch do

		name, texture, count, quality, canUse, level, minBid, minIncrement, buyoutPrice, bidAmount, highBidder, owner = GetAuctionItemInfo("list", index);

		-- Returns a number between 1 and 4
		-- 1 = short time (less then 30 min.)
		-- 2 = medium time (30 min. till 2 hours)
		-- 3 = long time (2 till 12 hours)
		-- 4 = very long time (more then 12 hours)
		timeleft = GetAuctionItemTimeLeft("list", index);
		if ( not timeleft or timeleft == 0 ) then
			timeleft = 4;
		end

		itemLink = GetAuctionItemLink("list", index);
		--_, _, _, itemLevel, _, itemType, itemSubType, _, itemEquipLoc = GetItemInfo(itemLink);

		itemId, itemMods, uId = MarketWatcher.DecodeItemLink(itemLink);

		if ( owner == "" ) then
			owner = nil;
		end

		if ( MarketWatcher.IsItemInTable(itemId) and not scan.processed[itemId] ) then
			MarketWatcher.SetTempTimestamp(itemId, status.queryTime);

			-- add auction to temp table; if seller name is blank, fill it with auction index so we can replace it with the correct name in future updates
			MarketWatcher.AddTempAuction(itemId, itemMods, uId, count, (owner or index), buyoutPrice, minBid, bidAmount, timeleft);
		end

		if ( not owner ) then
			status.waitForNames = true;
		end
	end
end

function scan.GetSellerNames()

	local batch, total = GetNumAuctionItems("list");

	local seller, owner;
	local itemLink, itemId;
	local noNilNames = true;
	
	for index = 1, batch do

		_, _, _, _, _, _, _, _, _, _, _, owner = GetAuctionItemInfo("list", index);

		itemLink = GetAuctionItemLink("list", index);
		itemId = MarketWatcher.DecodeItemLink(itemLink);

		if ( MarketWatcher.IsItemInTable(itemId) ) then

			for auctionIndex = 1, MarketWatcher.GetNumAuctions(itemId, 0) do

				_, seller = MarketWatcher.GetAuction(itemId, 0, auctionIndex);

				-- if auction index matches index in temp auction, replace with seller name
				if ( type(seller) == "number" and seller == index and owner and owner ~= "" ) then
					MarketWatcher.EditAuction(itemId, 0, auctionIndex, nil, nil, nil, owner or "");
				end
			end

		end

		if ( owner == nil ) then
			noNilNames = false;
		end
	end

	return noNilNames;
end

function scan.OnPageComplete()
	local entries;

	for itemId in MarketWatcher.TempItemIterator() do

		entries = MarketWatcher.GetNumAuctions(itemId, 0);

		if ( entries > 0 and not scan.processed[itemId] ) then

			-- blank out auctions without owners, otherwise getting the next page will fill in the wrong seller name
			for auctionIndex = 1, entries do

				_, seller = MarketWatcher.GetAuction(itemId, 0, auctionIndex);

				if ( type(seller) == "number" ) then
					MarketWatcher.EditAuction(itemId, 0, auctionIndex, nil, nil, nil, "");
				end
			end
		end
	end
end

function scan.OnQueryComplete()

	local i, j, entries, seller, instances;

	for itemId in MarketWatcher.TempItemIterator() do

		entries = MarketWatcher.GetNumAuctions(itemId, 0);

		if ( entries > 0 and not scan.processed[itemId] ) then

			-- compress duplicate auctions
			i = 1;
			while i < entries do

				j = i + 1;
				while j <= entries do

					if ( MarketWatcher.IsAuctionEqual(itemId, 0, i, 0, j) ) then

						_, _, _, _, _, _, instances = MarketWatcher.GetAuction(itemId, 0, i);

						if ( instances ) then
							instances = instances + 1;
						else
							instances = 2;
						end
						
						MarketWatcher.EditAuction(itemId, 0, i, nil, nil, nil, nil, nil, nil, nil, nil, instances);
						MarketWatcher.DeleteTempAuction(itemId, j);

						entries = entries - 1;
						j = j - 1;

					end
					j = j + 1;

				end
				i = i + 1;

			end

			-- sort auctions
			MarketWatcher.SortTempAuctions(itemId);

			scan.processed[itemId] = true;
		end
	end
end

function scan.OnScanComplete()

	status.scan = false;
	status.endTime = time();
	timer.StopTimer();

	MarketWatcherScanFrameAbortButton:Hide();
	MarketWatcherScanFrameStatusText:SetText(MARKETWATCHER_SCANCOMPLETE);
	MarketWatcherScanFrameStatusText:Hide();
	MarketWatcherScanFrameStatusText2:Hide();
	MarketWatcherScanFrameScanButton:Enable();

	local elapsed = status.endTime - status.startTime;
	local elapsedMinutes = floor(elapsed / 60);
	local elapsedSeconds = floor(elapsed % 60);
	MarketWatcherScanFrameScanTimeText:SetFormattedText(MARKETWATCHER_SCANTIME2, date("%X"), elapsedMinutes, elapsedSeconds, status.pagesScanned);

	scan.ComputeStats();
	scan.WriteHistory();
	scan.DisplayStats();

	MarketWatcherScanFrameRecordCheckBox:SetChecked(false);
end

function scan.AddQuery(name, minLevel, maxLevel, invTypeIndex, classIndex, subclassIndex, qualityIndex)

	table.insert(queries, {
		["name"] = name,
		["minLevel"] = minLevel,
		["maxLevel"] = maxLevel,
		["invTypeIndex"] = invTypeIndex,
		["classIndex"] = classIndex,
		["subclassIndex"] = subclassIndex,
		["qualityIndex"] = qualityIndex,
	});
end

function scan.ComputeStats()

	local function StackType(table, type, bid, buyout, instances, bidPerUnit, buyoutPerUnit, count)
		if ( not table[type] ) then
			table[type] = {};
			table[type].auctions = 0;
		end

		local t = table[type];

		t.auctions = t.auctions + instances;

		if ( not t.lowestBid or t.lowestBid > bid ) then
			t.lowestBid = bid;
		end

		if ( (not t.lowestBidPerUnit or t.lowestBidPerUnit > bidPerUnit) ) then
			t.lowestBidPerUnit = bidPerUnit;
			t.lowestBidPerUnitCount = count;
		end

		if ( (not t.lowestBuyout or t.lowestBuyout > buyout) and buyout > 0 ) then
			t.lowestBuyout = buyout;
		end

		if ( (not t.lowestPerUnitBuyout or t.lowestPerUnitBuyout > buyoutPerUnit) and buyout > 0 ) then
			t.lowestPerUnitBuyout = buyoutPerUnit;
			t.lowestPerUnitBuyoutCount = count;
		end
	end

	local itemTable, auction, itemInfo;
	local t, fullStacks, bidPerUnit, lowestBidPerUnit, lowestBuyoutPerUnit;

	for itemId in MarketWatcher.TempItemIterator() do

		t = {};
		itemInfo = MarketWatcher.GetItemInfoTable(itemId);

		t.itemId = itemId;
		t.totalAuctions = 0;
		t.totalForSale = 0;

		lowestBidPerUnit = 0;
		lowestBuyoutPerUnit = 0;

		for count, seller, buyout, buyoutPerUnit, currentBid, timeleft, instances, minBid, bidAmount in MarketWatcher.AuctionsIterator(itemId, 0) do

			t.totalAuctions = t.totalAuctions + instances;
			t.totalForSale = t.totalForSale + count * instances;

			bidPerUnit = math.ceil(currentBid / count);

			if ( not itemInfo.summaryStackOnly or (itemInfo.summaryStackOnly and itemInfo.stackCount == count) ) then

				if ( lowestBidPerUnit == 0 or bidPerUnit < lowestBidPerUnit ) then
					lowestBidPerUnit = bidPerUnit;
				end
				
				if ( buyout > 0 and (lowestBuyoutPerUnit == 0 or buyoutPerUnit < lowestBuyoutPerUnit) ) then
					lowestBuyoutPerUnit = buyoutPerUnit;
				end
			end

			if ( count == 1 ) then
				StackType(t, "single", currentBid, buyout, instances, bidPerUnit, buyoutPerUnit, count);
			else
				StackType(t, "nonSingle", currentBid, buyout, instances, bidPerUnit, buyoutPerUnit, count);
			end
			if ( itemInfo.stackCount == count ) then
				StackType(t, "fullStack", currentBid, buyout, instances, bidPerUnit, buyoutPerUnit, count);
			else
				StackType(t, "nonFullStack", currentBid, buyout, instances, bidPerUnit, buyoutPerUnit, count);
			end
		end

		t.lowestBuyoutPerUnit = lowestBuyoutPerUnit;
		t.lowestBidPerUnit = lowestBidPerUnit;

		-- if more than 2 full stack auctions, or fullstack auctions > single auctions, then just display fullStack and nonFullStack auction types and vice versa
		if ( t["fullStack"] and t["single"] and (t["fullStack"].auctions > 2 or t["fullStack"].auctions > t["single"].auctions) ) then
			t["single"] = nil;
			t["nonSingle"] = nil;
		end

		if ( t["single"] ) then
			t["fullStack"] = nil;
			t["nonFullStack"] = nil;
		end
		if ( t["fullStack"] ) then
			t["single"] = nil;
			t["nonSingle"] = nil;
		end

		if ( itemInfo.summaryStackOnly ) then
			t["single"] = nil;
			t["nonSingle"] = nil;
			t["nonFullStack"] = nil;
		end

		local display = true;

		if ( not itemInfo.noneOnSale and t.totalAuctions == 0 ) then
			display = false;

			if ( MarketWatcher.debug ) then
				print(itemInfo.name.." not on sale; not displaying");
			end
		end

		if ( display and itemInfo.summaryConditional == "below" ) then

			-- if a threshold is set and there are lowest bids/buyouts; display if no threshold is set even if below is selected
			if ( (itemInfo.bidThreshold > 0 and lowestBidPerUnit > 0) or (itemInfo.buyoutThreshold > 0 and lowestBuyoutPerUnit > 0) ) then
				
				display = false;

				-- if bid threshold is set and met
				if ( itemInfo.bidThreshold > 0 and lowestBidPerUnit > 0 and lowestBidPerUnit <= itemInfo.bidThreshold ) then
					display = true;
				end
			
				if ( itemInfo.buyoutThreshold > 0 and lowestBuyoutPerUnit > 0 and lowestBuyoutPerUnit <= itemInfo.buyoutThreshold ) then
					display = true;
				end

				if ( not display and MarketWatcher.debug ) then
					print("threshold not met for "..itemInfo.name);
				end
			end
		end

		if ( display and itemInfo.summaryConditional == "above" ) then

			-- if a threshold is set and there are lowest bids/buyouts; display if no threshold is set even if below is selected
			if ( (itemInfo.bidThreshold > 0 and lowestBidPerUnit > 0) or (itemInfo.buyoutThreshold > 0 and lowestBuyoutPerUnit > 0) ) then
				
				display = false;

				-- if bid threshold is set and met
				if ( itemInfo.bidThreshold > 0 and lowestBidPerUnit > 0 and lowestBidPerUnit >= itemInfo.bidThreshold ) then
					display = true;
				end
			
				if ( itemInfo.buyoutThreshold > 0 and lowestBuyoutPerUnit > 0 and lowestBuyoutPerUnit >= itemInfo.buyoutThreshold ) then
					display = true;
				end

				if ( not display and MarketWatcher.debug ) then
					print("threshold not met for "..itemInfo.name);
				end
			end
		end

		if ( display ) then
			table.insert(scanStats, t);
		end
	end

	table.sort(scanStats, function(a, b)
		local aQuality = MarketWatcher.GetItemInfoTable(a.itemId).quality;
		local bQuality = MarketWatcher.GetItemInfoTable(b.itemId).quality;

		local aName = MarketWatcher.GetItemInfoTable(a.itemId).name;
		local bName = MarketWatcher.GetItemInfoTable(b.itemId).name;

		if ( aQuality == bQuality ) then
			return aName < bName;
		else
			return aQuality > bQuality;
		end
	end);
end

function scan.WriteHistory()

	local saved;

	if ( MarketWatcherScanFrameRecordCheckBox:GetChecked() ) then

		for itemId in MarketWatcher.TempItemIterator() do

			if ( MarketWatcher.GetItemInfoTable(itemId).recordHistory ) then

				MarketWatcher.SaveTempItemData(itemId);
				saved = true;
			end

			MarketWatcher.PruneHistory(itemId);	-- delete entries exceeding max age or max entries
		end
	end

	if ( saved ) then
		MarketWatcher.GetConfig().lastRecordedScan = time();
		DEFAULT_CHAT_FRAME:AddMessage(format(MARKETWATCHER_SCAN_RECORDED, date("%X")), 0, 1, 1, 1);
	end
end

function scan.DisplayStats()

	local n, button, icon, textFontString, name, itemInfo, color, stats, text;
	local offset = FauxScrollFrame_GetOffset(MarketWatcherScanFrameStatsScrollFrame);

	local function BidThresholdColor(price)
		if ( (itemInfo.bidThreshold > 0 and price > itemInfo.bidThreshold) ) then
			return "|cff9e5252";

		elseif ( (itemInfo.bidThreshold > 0 and price <= itemInfo.bidThreshold) ) then
			return "|cff4a9c4a";
		else
			return "";
		end
	end
	local function BuyoutThresholdColor(price)
		if ( (itemInfo.buyoutThreshold > 0 and price > itemInfo.buyoutThreshold) ) then
			return "|cff9e5252";

		elseif ( (itemInfo.buyoutThreshold > 0 and price <= itemInfo.buyoutThreshold) ) then
			return "|cff4a9c4a";
		else
			return "";
		end
	end

	for i = 1, 8 do

		n = i + (offset * 2);

		button = _G["MarketWatcherScanFrameStatsButton"..i];
		icon = _G["MarketWatcherScanFrameStatsButton"..i.."ItemIconTexture"];
		name = _G["MarketWatcherScanFrameStatsButton"..i.."ItemName"];
		textFontString = _G["MarketWatcherScanFrameStatsButton"..i.."StatsText"];

		if ( scanStats[n] ) then
			stats = scanStats[n];
			itemInfo = MarketWatcher.GetItemInfoTable(scanStats[n].itemId);
			if ( not itemInfo ) then
				itemInfo = { texture = "", name = "", quality = 1 };
			end

			icon:SetTexture(itemInfo.texture);

			name:SetText(itemInfo.name);

			color = ITEM_QUALITY_COLORS[itemInfo.quality];
			name:SetVertexColor(color.r, color.g, color.b);

			text = format(MARKETWATCHER_TOTALS, stats.totalAuctions, stats.totalForSale);
			
			if ( stats.single ) then
				text = text..format(MARKETWATCHER_SINGLEAUCTIONS, stats.single.auctions);
				text = text..BidThresholdColor(stats.single.lowestBidPerUnit);
				text = text..format(MARKETWATCHER_LOWESTBID, MarketWatcher.MoneyText(stats.single.lowestBid)).."\n";
				if ( stats.single.lowestBuyout ) then
					text = text..BuyoutThresholdColor(stats.single.lowestBuyout);
					text = text..format(MARKETWATCHER_LOWESTBUYOUT, MarketWatcher.MoneyText(stats.single.lowestBuyout));
					if ( not stats.single.historical ) then
						stats.single.historical = scan.HistoricalComparison(stats.itemId, stats.single.lowestBuyout);
					end
					text = text.." "..stats.single.historical.."\n";
				end

				if ( stats.nonSingle ) then
					text = text..format(MARKETWATCHER_NONSINGLEAUCTIONS, stats.nonSingle.auctions);
					if ( stats.nonSingle.lowestBuyout ) then
						text = text..BuyoutThresholdColor(stats.nonSingle.lowestPerUnitBuyout);
						text = text..format(MARKETWATCHER_LOWESTBUYOUTPERUNIT, MarketWatcher.MoneyText(stats.nonSingle.lowestPerUnitBuyout)).." (x"..stats.nonSingle.lowestPerUnitBuyoutCount..")";
						if ( not stats.nonSingle.historical ) then
							stats.nonSingle.historical = scan.HistoricalComparison(stats.itemId, stats.nonSingle.lowestPerUnitBuyout);
						end
						text = text.." "..stats.nonSingle.historical;
					end
				end

			elseif ( stats.fullStack ) then
				text = text..format(MARKETWATCHER_FULLSTACKAUCTIONS, stats.fullStack.auctions);
				text = text..BidThresholdColor(stats.fullStack.lowestBidPerUnit);
				text = text..format(MARKETWATCHER_LOWESTBID, MarketWatcher.MoneyText(stats.fullStack.lowestBid));
				if ( stats.fullStack.lowestBuyout ) then
					text = text.."  "..format(MARKETWATCHER_LOWESTBUYOUT, MarketWatcher.MoneyText(stats.fullStack.lowestBuyout)).."\n";
					text = text..BuyoutThresholdColor(stats.fullStack.lowestPerUnitBuyout);
					text = text..format(MARKETWATCHER_LOWESTBUYOUTPERUNIT, MarketWatcher.MoneyText(stats.fullStack.lowestPerUnitBuyout));
					if ( not stats.fullStack.historical ) then
						stats.fullStack.historical = scan.HistoricalComparison(stats.itemId, stats.fullStack.lowestPerUnitBuyout);
					end
					text = text.." "..stats.fullStack.historical.."\n";
				else
					text = text.."\n";
				end
			end
			if ( not stats.single and stats.nonFullStack ) then
				text = text..format(MARKETWATCHER_NONFULLSTACKAUCTIONS, stats.nonFullStack.auctions);
				if ( stats.nonFullStack.lowestBuyout ) then
					text = text..BuyoutThresholdColor(stats.nonFullStack.lowestPerUnitBuyout);
					text = text..format(MARKETWATCHER_LOWESTBUYOUTPERUNIT, MarketWatcher.MoneyText(stats.nonFullStack.lowestPerUnitBuyout)).." (x"..stats.nonFullStack.lowestPerUnitBuyoutCount..")";
					if ( not stats.nonFullStack.historical ) then
						stats.nonFullStack.historical = scan.HistoricalComparison(stats.itemId, stats.nonFullStack.lowestPerUnitBuyout);
					end
					text = text.." "..stats.nonFullStack.historical;
				end
			end

			textFontString:SetText(text);
			button.n = n;
			button.itemId = scanStats[n].itemId;

			if ( scan.selectedItem and scan.selectedItem == button.itemId ) then
				button:LockHighlight();
			else
				button:UnlockHighlight();
			end

			button:Show();
		else
			button:Hide();
		end
	end

	FauxScrollFrame_Update(MarketWatcherScanFrameStatsScrollFrame, math.ceil(#scanStats / 2), 4, 85);
end

-- returns "n%" in a shade of green (-50) to yellow (0) to red (100+)
local function ColorizePct(n)
	local red, green, blue = 255, 255, "00";

	if ( n > 0 ) then
		green = n * 2.5;
		if ( green > 255 ) then
			green = 255;
		end
		green = 255 - green;
		return "|cff"..format("%.2x", red)..format("%.2x", green)..blue.."+"..n.."%|r";
	else
		red = math.abs(n) * 5;
		if ( red > 255 ) then
			red = 255;
		end
		red = 255 - red;
		return "|cff"..format("%.2x", red)..format("%.2x", green)..blue..n.."%|r";
	end
end

function scan.HistoricalComparison(itemId, pricePerUnit)

	local numScans = MarketWatcher.GetNumScans(itemId);
	if ( numScans < 2 ) then
		return "";
	end

	local timestamp, buyout, count, instances, units, timeStamp;
	local high, low, weeklyAverage, weeklyHighestLow, weeklyLowestLow, monthlyAverage, monthlyHighestLow, monthlyLowestLow;
	local monthlyAverageLow, weeklyAverageLow, weeklyScans, monthlyScans = 0, 0, 0, 0;
	local weekAgo = time() - 604800;
	local monthAgo = time() - 2419200;

	for scanIndex = numScans, 1, -1 do
		timestamp = MarketWatcher.GetScanTimestamp(itemId, scanIndex);
		units = 0;
		average = 0;
		low = 0;
		high = 0;

		low, high, average, _, _, _, _, _, units, timeStamp = MarketWatcher.ScanAnalysis(itemId, scanIndex);

		if ( low ) then
			if ( timeStamp < monthAgo ) then
				break;
			end

			if ( timeStamp > weekAgo ) then
				if ( not weeklyHighestLow or weeklyHighestLow < low ) then
					weeklyHighestLow = low;
				end
				if ( not weeklyLowestLow or weeklyLowestLow > low ) then
					weeklyLowestLow = low;
				end
				weeklyAverageLow = weeklyAverageLow + low;
				weeklyScans = weeklyScans + 1;
			end

			if ( not monthlyHighestLow or monthlyHighestLow < low ) then
				monthlyHighestLow = low;
			end
			if ( not monthlyLowestLow or monthlyLowestLow > low ) then
				monthlyLowestLow = low;
			end

			monthlyAverageLow = monthlyAverageLow + low;
			monthlyScans = monthlyScans + 1;
		end
	end

	weeklyAverageLow = weeklyAverageLow / weeklyScans;
	monthlyAverageLow = monthlyAverageLow / monthlyScans;

	local text = "";

	if ( weeklyScans > 0 ) then
		low = pricePerUnit - weeklyAverageLow;
		low = math.floor(100 / weeklyAverageLow * low);

		text = text..ColorizePct(low);
	end
	if ( monthlyScans > 0 and monthlyScans > weeklyScans ) then
		low = pricePerUnit - monthlyAverageLow;
		low = math.floor(100 / monthlyAverageLow * low);

		if ( weeklyScans > 0 ) then
			text = text.."/";
		end
		text = text..ColorizePct(low);
	end

	return text;
end

function scan.SetMouseoverFrame(button)
	if ( not scanStats or #scanStats == 0 ) then
		return
	end

	if ( button.n and scanStats[button.n] ) then

		local frame = MarketWatcherScanMouseoverFrame;
		local fontString = _G[frame:GetName().."Text"];
		local stats = scanStats[button.n];
		local itemId = stats.itemId;
		local itemInfo = MarketWatcher.GetItemInfoTable(itemId);
		local _, _, _, qualityHex = GetItemQualityColor(itemInfo.quality);
		local found;
		local list = {};

		local text = "|T"..itemInfo.texture..":0:0:0:-1|t"..qualityHex.." "..itemInfo.name.."|r\n";
		text = text..AUCTION_TOOLTIP_BUYOUT_PREFIX:sub(1, -2).."\n\n"

		for count, seller, buyout, perUnitBuyout, bid, timeleft, instances in MarketWatcher.AuctionsIterator(itemId, 0) do

			found = false;
			for _, v in ipairs(list) do
				if ( v[2] == count and v[3] == perUnitBuyout ) then
					v[1] = v[1] + instances;
					found = true;
					break;
				end
			end
			
			if ( not found ) then
				tinsert(list, { instances, count, perUnitBuyout, });
			end
		end

		table.sort(list, function(a, b)
			if ( a[3] == 0 ) then
				return false;
			end
			if ( b[3] == 0 ) then
				return true;
			end

			if ( a[3] == b[3] ) then
				return a[2] < b[2];
			else
				return a[3] < b[3];
			end
		end);
		for i, v in ipairs(list) do

			if ( v[1] > 1 ) then
				text = text.."|cff00ffff"..v[1].."|rx ";
			else
				text = text.."     ";
			end
			if ( v[2] < 10 and v[1] < 10 ) then
				text = text.."  ";
			end
			if ( v[3] > 0 ) then
				text = text.."[|cffffffff"..v[2].."|r] "..MarketWatcher.MoneyText(v[3]).."\n";
			else
				text = text.."[|cffffffff"..v[2].."|r] --\n";
			end

			list[i][1] = nil;
			list[i][2] = nil;
			list[i][3] = nil;
			list[i] = nil;
		end
		fontString:SetText(text);
		frame:SetHeight(fontString:GetStringHeight() + 10);
		frame:SetWidth(fontString:GetStringWidth() + 12);
	else
		frame:Hide();
	end
end

function MarketWatcherScanStatsButton_OnClick(self, button)
	if ( button == "LeftButton" ) then
		scan.selectedItem = self.itemId;
		scan.DisplayStats();

		MarketWatcher.listFunction = scan.ListAuctions_Update;
		scan.page = 0;
		MarketWatcher.QueryItem(scan.selectedItem, scan.page);

	elseif ( button == "RightButton" ) then
		
	end
end

function scan.ListAuctions_Update()
	local numBatchAuctions, totalAuctions = GetNumAuctionItems("list");
	local offset = FauxScrollFrame_GetOffset(MarketWatcherScanFrameAuctionsScrollFrame);

	if ( scan.selectedItem ) then
		MarketWatcherScanFrameAuctions:Show();
	else
		MarketWatcherScanFrameAuctions:Hide();
		return
	end

	MarketWatcherScanFrameBuyoutButton:Disable();

	local itemInfo = MarketWatcher.GetItemInfoTable(scan.selectedItem);
	local itemId, itemLink, auctions, button, buttonName, name, count, seller, buyoutPrice, buyoutPricePerUnit, minBid, minIncrement, bidAmount, highBidder, timeleft, instances, texture;
	local sortedAuctions = {};
	local NUM_BUTTONS = 20;

	for i = 1, numBatchAuctions do
		name, texture, count, _, _, _, minBid, minIncrement, buyoutPrice, bidAmount, highBidder, seller = GetAuctionItemInfo("list", i);
		timeleft = GetAuctionItemTimeLeft("list", i);
		itemLink = GetAuctionItemLink("list", i);
		itemId = MarketWatcher.DecodeItemLink(itemLink);
			
		if ( buyoutPrice == 0 ) then
			if ( bidAmount == 0 ) then
				bidAmount = minBid;
			else
				bidAmount = bidAmount + minIncrement;
			end
		else
			bidAmount = buyoutPrice;
		end

		buyoutPricePerUnit = ceil(buyoutPrice / count);
		if ( itemId == scan.selectedItem ) then
			table.insert( sortedAuctions, { count, seller, timeleft, bidAmount, buyoutPrice, buyoutPricePerUnit, highBidder, texture, i } );
		end
	end

	auctions = #sortedAuctions;
	if ( offset > (auctions - NUM_BUTTONS) ) then
		MarketWatcherScanFrameAuctionsScrollFrameScrollBar:SetValue(0);
		offset = 0;
	end

	table.sort(sortedAuctions, function(a, b)
		-- if buyout price == 0, push it below all auctions with buyouts
		if ( a[5] == 0 ) then
			return false;
		end
		if ( b[5] == 0 ) then
			return true;
		end

		if ( a[6] == b[6] ) then
			return a[1] < b[1];
		else
			return a[6] < b[6];
		end
	end);

	if ( numBatchAuctions == 0 ) then
		MarketWatcherScanFrameAuctionsText:Show();
		MarketWatcherScanFrameAuctionsText:SetText(BROWSE_NO_RESULTS);
	else
		if ( auctions == 0 ) then
			MarketWatcherScanFrameAuctionsText:Show();
			MarketWatcherScanFrameAuctionsText:SetFormattedText(MARKETWATCHER_ITEM_NOT_IN_PAGE, itemInfo.name);
		else
			MarketWatcherScanFrameAuctionsText:Hide();
		end
	end


	for i, auction in ipairs(sortedAuctions) do
		sortedAuctions[i] = nil;

		if ( i > offset and i <= (offset + NUM_BUTTONS) ) then

			buttonName = "MarketWatcherScanFrameAuctionsButton"..(i - offset);
			button = _G[buttonName];

			_G[buttonName.."Seller"]:SetText(auction[2]);
			_G[buttonName.."TimeLeft"]:SetText(AuctionFrame_GetTimeLeftText(auction[3]));
			if ( auction[5] == 0 ) then
				_G[buttonName.."CostPerUnit"]:SetText(MARKETWATCHER_NOBUYOUT);
				_G[buttonName.."Cost"]:SetText(MarketWatcher.MoneyText(auction[4]));
			else
				_G[buttonName.."Cost"]:SetText(MarketWatcher.MoneyText(auction[5]));
				_G[buttonName.."CostPerUnit"]:SetText(MarketWatcher.MoneyText(auction[6]));
			end
			_G[buttonName.."ItemIconTexture"]:SetTexture(auction[8]);
			if ( auction[1] > 1 ) then
				_G[buttonName.."ItemCount"]:SetText(auction[1]);
				_G[buttonName.."ItemCount"]:Show();
			else
				_G[buttonName.."ItemCount"]:Hide();
			end

			if ( GetSelectedAuctionItem("list") and GetSelectedAuctionItem("list") == auction[9] ) then
				button:LockHighlight();

				if ( not auction[7] and auction[5] > 0 and auction[2] ~= UnitName("player") ) then
					MarketWatcherScanFrameBuyoutButton:Enable();
				end
			else
				button:UnlockHighlight();
			end

			button.index = auction[9];
			button:Show();
		end
	end

	-- if some buttons need to be hidden
	if ( (offset + NUM_BUTTONS) > auctions ) then
		for i = ((auctions - offset) + 1), NUM_BUTTONS do
			button = _G["MarketWatcherScanFrameAuctionsButton"..i];
			if ( button ) then
				button:Hide();
			end
		end
	end

	local lastPage = ceil(totalAuctions / NUM_AUCTION_ITEMS_PER_PAGE) - 1;
	MarketWatcherScanFrameAuctionsSearchCountText:SetFormattedText(MERCHANT_PAGE_NUMBER, scan.page + 1, lastPage + 1);
	MarketWatcherScanFrameAuctionsSearchCountText:Show();

	if ( totalAuctions > NUM_AUCTION_ITEMS_PER_PAGE ) then
		MarketWatcherScanFrameAuctionsPrevPageButton:Show();
		MarketWatcherScanFrameAuctionsNextPageButton:Show();
		--local itemsMin = AuctionFrameBrowse.page * NUM_AUCTION_ITEMS_PER_PAGE + 1;
		--local itemsMax = itemsMin + numBatchAuctions - 1;
		--MarketWatcherScanFrameAuctionsSearchCountText:SetFormattedText(NUMBER_OF_RESULTS_TEMPLATE, itemsMin, itemsMax, totalAuctions);
		if ( scan.page == 0 ) then
			MarketWatcherScanFrameAuctionsPrevPageButton:Disable();
		else
			MarketWatcherScanFrameAuctionsPrevPageButton:Enable();
		end
		if ( scan.page == lastPage ) then
			MarketWatcherScanFrameAuctionsNextPageButton:Disable();
		else
			MarketWatcherScanFrameAuctionsNextPageButton:Enable();
		end
		
	else
		MarketWatcherScanFrameAuctionsPrevPageButton:Hide();
		MarketWatcherScanFrameAuctionsNextPageButton:Hide();

		if ( totalAuctions == 0 ) then
			MarketWatcherScanFrameAuctionsSearchCountText:Hide();
		end
	end

	FauxScrollFrame_Update(MarketWatcherScanFrameAuctionsScrollFrame, auctions, NUM_BUTTONS, 16);
end

function scan.AuctionButton_OnClick(self, button)
	if ( button == "LeftButton" ) then
		SetSelectedAuctionItem("list", self.index);
		scan.ListAuctions_Update();

	elseif ( button == "RightButton" and IsShiftKeyDown() ) then
		local name, _, _, _, _, _, _, _, buyoutPrice = GetAuctionItemInfo("list", self.index);
		PlaceAuctionBid("list", self.index, buyoutPrice);
		self:Hide();
		GameTooltip:Hide();
	end
end

function scan.ListQuery()

	local itemInfo;
	local invTypeIndex, classIndex, subclassIndex;

	itemInfo = MarketWatcher.GetItemInfoTable(scan.selectedItem);

	classIndex, subclassIndex = MarketWatcher.GetItemAuctionClasses(itemInfo.type, itemInfo.subType);
	invTypeIndex = MarketWatcher.GetItemInvType(itemInfo.equipLoc, classIndex, subclassIndex);

	MarketWatcher.QueryServer(itemInfo.name, nil, nil, invTypeIndex, classIndex, subclassIndex, scan.page, 0, itemInfo.quality, false);
end

function scan.AuctionButton_OnEnter(self)
	GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
	
	local name, texture, count, quality, canUse, level, minBid, minIncrement, buyoutPrice, bidAmount, highBidder, owner = GetAuctionItemInfo("list", self.index);

	if ( bidAmount > 0 ) then
		minBid = bidAmount + minIncrement;
	end

	local color = { GetItemQualityColor(quality) };
	GameTooltip:AddLine( name, color[1], color[2], color[3] );

	GameTooltip:AddLine("|n");
	SetTooltipMoney(GameTooltip, ceil(minBid / count), "STATIC", AUCTION_TOOLTIP_BID_PREFIX);
	if ( buyoutPrice > 0 ) then
		SetTooltipMoney(GameTooltip, ceil(buyoutPrice / count), "STATIC", AUCTION_TOOLTIP_BUYOUT_PREFIX);
	end

	GameTooltip:Show();
	GameTooltip_ShowCompareItem();
end
